@@ -37,6 +37,10 @@ struct module;
 #define AUTH_MASTER_WAITING_MSG \
 	"Waiting for authentication master process to respond.."
 
+/* Client logged out without having successfully authenticated. */
+#define CLIENT_UNAUTHENTICATED_LOGOUT_MSG \
+	"Aborted login by logging out"
+
 struct master_service_connection;
 
 enum client_disconnect_reason {
@@ -77,6 +81,18 @@ enum client_auth_result {
 	CLIENT_AUTH_RESULT_ANONYMOUS_DENIED
 };
 
+enum client_list_type {
+	CLIENT_LIST_TYPE_NONE = 0,
+	/* clients (disconnected=FALSE, fd_proxying=FALSE, destroyed=FALSE) */
+	CLIENT_LIST_TYPE_ACTIVE,
+	/* destroyed_clients (destroyed=TRUE, fd_proxying=FALSE). Either the
+	   client will soon be freed or it's only referenced via
+	   "login_proxies". */
+	CLIENT_LIST_TYPE_DESTROYED,
+	/* client_fd_proxies (fd_proxying=TRUE) */
+	CLIENT_LIST_TYPE_FD_PROXY,
+};
+
 struct client_auth_reply {
 	const char *master_user, *reason;
 	enum client_auth_fail_code fail_code;
@@ -87,12 +103,14 @@ struct client_auth_reply {
 	in_port_t port;
 	unsigned int proxy_timeout_msecs;
 	unsigned int proxy_refresh_secs;
+	unsigned int proxy_host_immediate_failure_after_secs;
 	enum login_proxy_ssl_flags ssl_flags;
 
 	/* all the key=value fields returned by passdb */
 	const char *const *all_fields;
 
 	bool proxy:1;
+	bool proxy_noauth:1;
 	bool proxy_nopipelining:1;
 	bool proxy_not_trusted:1;
 	bool nologin:1;
@@ -123,7 +141,9 @@ struct client_vfuncs {
 			    const char *text);
 	void (*proxy_reset)(struct client *client);
 	int (*proxy_parse_line)(struct client *client, const char *line);
-	void (*proxy_error)(struct client *client, const char *text);
+	void (*proxy_failed)(struct client *client,
+			     enum login_proxy_failure_type type,
+			     const char *reason, bool reconnecting);
 	const char *(*proxy_get_state)(struct client *client);
 	void (*send_raw_data)(struct client *client,
 			      const void *data, size_t size);
@@ -132,11 +152,9 @@ struct client_vfuncs {
 };
 
 struct client {
-	/* If disconnected=FALSE, the client is in "clients" list.
-	   If fd_proxying=TRUE, the client is in "client_fd_proxies" list.
-	   Otherwise, either the client will soon be freed or it's only
-	   referenced via "login_proxies" which doesn't use these pointers. */
 	struct client *prev, *next;
+	/* Specifies which linked list the client is in */
+	enum client_list_type list_type;
 
 	pool_t pool;
 	/* this pool gets free'd once proxying starts */
@@ -144,7 +162,7 @@ struct client {
 	struct client_vfuncs v;
 	struct client_vfuncs *vlast;
 
-	time_t created;
+	struct timeval created;
 	int refcount;
 	struct event *event;
 
@@ -156,6 +174,7 @@ struct client {
 	struct ssl_iostream *ssl_iostream;
 	const struct login_settings *set;
 	const struct master_service_ssl_settings *ssl_set;
+	const struct master_service_ssl_server_settings *ssl_server_set;
 	const char *session_id, *listener_name, *postlogin_socket_path;
 	const char *local_name;
 	const char *client_cert_common_name;
@@ -181,11 +200,17 @@ struct client {
 	unsigned int proxy_ttl;
 
 	char *auth_mech_name;
+	enum sasl_server_auth_flags auth_flags;
 	struct auth_client_request *auth_request;
 	string_t *auth_response;
 	time_t auth_first_started, auth_finished;
 	const char *sasl_final_resp;
+	const char *sasl_final_delayed_resp;
+	enum sasl_server_reply delayed_final_reply;
+	const char *const *final_args;
 	const char *const *auth_passdb_args;
+	struct anvil_query *anvil_query;
+	struct anvil_request *anvil_request;
 
 	unsigned int master_auth_id;
 	unsigned int master_tag;
@@ -223,17 +248,20 @@ struct client {
 	bool ssl_servername_settings_read:1;
 	bool banner_sent:1;
 	bool authenticating:1;
+	bool auth_client_continue_pending:1;
 	bool auth_try_aborted:1;
 	bool auth_initializing:1;
 	bool auth_process_comm_fail:1;
 	bool auth_anonymous:1;
 	bool proxy_auth_failed:1;
+	bool proxy_noauth:1;
 	bool proxy_nopipelining:1;
 	bool proxy_not_trusted:1;
 	bool auth_waiting:1;
 	bool notified_auth_ready:1;
 	bool notified_disconnect:1;
 	bool fd_proxying:1;
+	bool final_response:1;
 	/* ... */
 };
 
@@ -258,10 +286,13 @@ struct client *
 client_alloc(int fd, pool_t pool,
 	     const struct master_service_connection *conn,
 	     const struct login_settings *set,
-	     const struct master_service_ssl_settings *ssl_set);
+	     const struct master_service_ssl_settings *ssl_set,
+	     const struct master_service_ssl_server_settings *ssl_server_set);
 void client_init(struct client *client, void **other_sets);
-void client_disconnect(struct client *client, const char *reason);
+void client_disconnect(struct client *client, const char *reason,
+		       bool add_disconnected_prefix);
 void client_destroy(struct client *client, const char *reason);
+void client_destroy_iostream_error(struct client *client);
 /* Destroy the client after a successful login. Either the client fd was
    sent to the post-login process, or the connection will be proxied. */
 void client_destroy_success(struct client *client, const char *reason);
@@ -310,6 +341,9 @@ void client_send_raw(struct client *client, const char *data);
 void client_common_send_raw_data(struct client *client,
 				 const void *data, size_t size);
 void client_common_default_free(struct client *client);
+void client_common_proxy_failed(struct client *client,
+				enum login_proxy_failure_type type,
+				const char *reason, bool reconnecting);
 
 void client_set_auth_waiting(struct client *client);
 void client_auth_send_challenge(struct client *client, const char *data);
@@ -318,16 +352,17 @@ int client_auth_begin(struct client *client, const char *mech_name,
 		      const char *init_resp);
 int client_auth_begin_private(struct client *client, const char *mech_name,
 			      const char *init_resp);
+int client_auth_begin_implicit(struct client *client, const char *mech_name,
+			       const char *init_resp);
 bool client_check_plaintext_auth(struct client *client, bool pass_sent);
 int client_auth_read_line(struct client *client);
 
 void client_proxy_finish_destroy_client(struct client *client);
 void client_proxy_log_failure(struct client *client, const char *line);
-void client_proxy_failed(struct client *client, bool send_line);
 const char *client_proxy_get_state(struct client *client);
 
 void clients_notify_auth_connected(void);
-void client_destroy_oldest(void);
+bool client_destroy_oldest(bool kill, struct timeval *created_r);
 void clients_destroy_all(void);
 void clients_destroy_all_reason(const char *reason);
 
